/*
 * block_io.h
 *
 * Copyright 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * Distributed under GPLv2.
 *
 * This file contains declarations for functions exported from
 * block_io.c, which contains low level io functions.
 */

#include <linux/suspend-common.h>

#define MAX_READAHEAD (int) sizeof(unsigned long) * 8
#ifdef MAX_READAHEAD
extern unsigned long suspend_readahead_flags[];
extern spinlock_t suspend_readahead_flags_lock;
extern void suspend_wait_on_readahead(int readahead);
#endif

// Forward Declarations

struct submit_params {
	swp_entry_t swap_address;
	struct page * page;
	DEVICE_BLOCK_TYPE dev;
	long blocks[PAGE_SIZE/512];
	int blocks_used;
	int readahead_index;
};


// Lowlevel I/O functions
// - Get/set block size.
int suspend_get_block_size(DEVICE_BLOCK_TYPE bdev);
int suspend_set_block_size(DEVICE_BLOCK_TYPE bdev, int size);

// - Manage statistics for I/O structures.
void __inline reset_io_stats(void);
void suspend_check_io_stats(void);

// - Submit & cleanup I/O.
void cleanup_completed_io(void);
void finish_all_io(void);

// - Synchronous (async + wait on completion)
extern void bdev_page_io(int rw, DEVICE_BLOCK_TYPE bdev, long pos,
		struct page * page);

extern int max_async_ios;
#define REAL_MAX_ASYNC ((max_async_ios ? max_async_ios : 32))

extern void do_suspend_io(int rw, 
		struct submit_params * submit_info, int syncio);

/* swap_entry_to_range_val & range_val_to_swap_entry: 
 * We are putting offset in the low bits so consecutive swap entries
 * make consecutive range values */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
extern dev_t name_to_dev_t(char *line) __init;
#define swap_entry_to_range_val(swp_entry) (swp_entry.val)
#define range_val_to_swap_entry(val) (swp_entry_t) { (val) }
#else
extern kdev_t name_to_kdev_t(char *line) __init;
#define swap_entry_to_range_val(swp_entry) ((swp_entry.val >> 8) | \
		((swp_entry.val & 0x3f) << 24))
#define range_val_to_swap_entry(val) (swp_entry_t) { ((val >> 24) | \
		((val & 0xffffff) << 8)) }
#endif
