/*
 * NFSwriter.c
 *
 * Copyright 2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * Distributed under GPLv2.
 * 
 * This file encapsulates functions for storing an image on
 * an NFS share.
 *
 */

#include <linux/suspend-common.h>

static int nfswriter_mount_share(void)
{
	void *data = nfs_root_data();
	
	sys_mkdir("/proc/swsusp/nfsroot", 0700);

	if (data && sys_mount("/dev/root","/proc/swsusp/nfsroot","nfs",root_mountflags,data) == 0)
		return 1;

	return 0;
}

static int nfswriter_unmount_share(void)
{
	sys_rmdir("/proc/swsusp/nfsroot");
}

static unsigned long nfswriter_storage_allocated(void)
{
	return 0;
}

static long nfswriter_storage_available(void)
{
	return (long) (0);
}

static int nfswriter_release_storage(void)
{
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if ((TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) && (now_resuming))
		return 0;
#endif

	/* Invalidate the image */
	return 0;
}

static long nfswriter_allocate_header_space(unsigned long space_requested)
{
}

static int nfswriter_allocate_storage(unsigned long space_requested)
{
}

static int nfswriter_write_header_init(void)
{
	return 0;
}

static int nfswriter_write_header_chunk(char * buffer, int buffer_size)
{
	return 0;
}

static int nfswriter_write_header_cleanup(void)
{
	return 0;
}

/* ------------------------- HEADER READING ------------------------- */

static int nfswriter_read_header_init(void)
{
	return 0;
}

static int nfswriter_read_header_chunk(char * buffer, int buffer_size)
{
	return buffer_size;
}

static int nfswriter_read_header_cleanup(void)
{
	return 0;
}

static int nfswriter_prepare_save_ranges(void)
{
	return 0;
}

static int nfswriter_post_load_ranges(void)
{
	return 0;
}

static int nfswriter_write_init(int stream_number)
{
	return 0;
}

static int nfswriter_write_chunk(char * buffer)
{
	return 0;
}

static int nfswriter_write_cleanup(void)
{
	return 0;
}

static int nfswriter_read_init(int stream_number)
{
	return 0;
}

static int nfswriter_read_chunk(char * buffer, int sync)
{
	return 0;
}

static int nfswriter_read_cleanup(void)
{
	return 0;
}

static int nfswriter_invalidate_image(void)
{
	return 0;
}

/*
 * workspace_size
 *
 * Description:
 * Returns the number of bytes of RAM needed for this
 * code to do its work. (Used when calculating whether
 * we have enough memory to be able to suspend & resume).
 *
 */
static unsigned long nfswriter_memory_needed(void)
{
	return 0;
}

/*
 * Storage needed
 *
 * Returns amount of space in the header required
 * for the nfswriter's data.
 *
 */
static unsigned long nfswriter_storage_needed(void)
{
	return 0;
}

/*
 * Wait on I/O
 *
 */

static int nfswriter_wait_on_io(int flush_all)
{
	return 0;
}

/*
 * Image_exists
 *
 */

static int nfswriter_image_exists(void)
{
	return 0;
}

/*
 * Parse Image Location
 *
 */

static int nfswriter_parse_image_location(char * commandline, int boot_time, 
		int only_writer)
{
	int posn = 0;
	
	if (strncmp(commandline, "nfs:", 5)) {
		if (!only_writer) {
			printk(name_suspend "Commandline doesn't begin with 'nfs:'"
					"... nfswriter ignoring.\n");
			return 0;
		}

	} else
		posn+= 5;
	
	/*
	 * posn now points to the start of the parameters we need to parse.
	 */

	return 1;
}

struct suspend_plugin_ops nfswriterops = {
	.type					= WRITER_PLUGIN,
	.name					= "NFS Writer",
	.memory_needed				= nfswriter_memory_needed,
	.storage_needed				= nfswriter_storage_needed,
	.write_init				= nfswriter_write_init,
	.write_chunk				= nfswriter_write_chunk,
	.write_cleanup				= nfswriter_write_cleanup,
	.read_init				= nfswriter_read_init,
	.read_chunk				= nfswriter_read_chunk,
	.read_cleanup				= nfswriter_read_cleanup,
	.ops = {
		.writer = {
			.storage_available 	= nfswriter_storage_available,
			.storage_allocated	= nfswriter_storage_allocated,
			.release_storage	= nfswriter_release_storage,
			.allocate_header_space	= 
				nfswriter_allocate_header_space,
			.allocate_storage	= nfswriter_allocate_storage,
			.image_exists		= nfswriter_image_exists,
			.write_header_init	= nfswriter_write_header_init,
			.write_header_chunk	= nfswriter_write_header_chunk,
			.write_header_cleanup	= 
				nfswriter_write_header_cleanup,
			.read_header_init	= nfswriter_read_header_init,
			.read_header_chunk	= nfswriter_read_header_chunk,
			.read_header_cleanup	= 
				nfswriter_read_header_cleanup,
			.prepare_save_ranges	= 
				nfswriter_prepare_save_ranges,
			.post_load_ranges	= nfswriter_post_load_ranges,
			.invalidate_image	= nfswriter_invalidate_image,
			.wait_on_io		= nfswriter_wait_on_io,
			.parse_image_location	= 
				nfswriter_parse_image_location,
		}
	}
};

/* ---- Registration ---- */

static __init int nfswriter_load(void)
{
	printk("Software Suspend NFS Writer v0.1\n");
	return suspend_register_plugin(&nfswriterops);
}

__initcall(nfswriter_load);

